!Module for solving one step of the elastic wave equation
module Prop_elastic_step


  use D10pt_mod
  use S_mod

  implicit none
  !Module variables
  integer, private								:: nz, nx, ny, nt
  real, private									:: dt
  real,allocatable,dimension(:,:,:), private	:: scalevz, scalevx, scale_lame_mu, scale_lame, scale_mu_xz, ptemp
  

  contains
  
  	subroutine Prop_elastic_step_init(nz_in, nx_in, ny_in, nt_in, nv_pad_in, dt_in, dz_in, dx_in, dy_in, twoD_in, rhox_in, rhoz_in, lame_in, mu_in)
  		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in, nt_in, nv_pad_in 
		real									:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in
		real, dimension(:,:,:), pointer			:: lame_in, mu_in
		real, dimension(:,:,:), allocatable		:: rhox_in, rhoz_in
		nz = nz_in
		nx = nx_in
		ny = ny_in
		nt = nt_in
		dt = dt_in
		
		
		!Initializing other modules
		call D10pt_init(nz,nx,ny,dz_in,dx_in,dy_in,twoD_in)
		call S_init(nz,nx,ny,twoD_in)
		
		if(.not. allocated(scalevz)) allocate(scalevz(nz,nx,ny))
		if(.not. allocated(scalevx)) allocate(scalevx(nz,nx,ny))
		if(.not. allocated(scale_lame_mu)) allocate(scale_lame_mu(nz,nx,ny))
		if(.not. allocated(scale_lame)) allocate(scale_lame(nz,nx,ny))
		if(.not. allocated(scale_mu_xz)) allocate(scale_mu_xz(nz,nx,ny))
		if(.not. allocated(ptemp)) allocate(ptemp(nz,nx,ny))
		
		!Computing the scaling factors of the derivatives
		scalevx = dt/rhox_in
		scalevz = dt/rhoz_in
		call Sx_op(.true., .false., scale_lame, dt*mu_in)
  		call Sz_op(.true., .false., scale_mu_xz, scale_lame)
  		scale_lame = dt*lame_in
  		scale_lame_mu = dt*(lame_in+2.0*mu_in)
		
  
  	end subroutine Prop_elastic_step_init
  
  
  	subroutine Prop_elastic_step_op(adj, pcur_xx, pcur_zz, pcur_xz, pcur_vx, pcur_vz, pxx_dt, pzz_dt,pxz_dt, pvx_dt, pvz_dt)
  		!Input-Output variables
		logical, intent(in)								:: adj
		real, dimension(nz,nx,ny), intent(inout)		:: pcur_xx, pcur_zz, pcur_xz, pcur_vx, pcur_vz
		real, dimension(nz,nx,ny), intent(inout)		:: pxx_dt, pzz_dt,pxz_dt, pvx_dt, pvz_dt
		
		
		
		
		if(adj) then !Adjoint 
			!vx time derivative
			ptemp=pcur_xx*scale_lame_mu
			call Dpx_op(.true.,.false.,pvx_dt,ptemp)  
			ptemp=pcur_zz*scale_lame
			call Dpx_op(.true.,.true.,pvx_dt,ptemp) 
			ptemp=pcur_xz*scale_mu_xz
			call Dmz_op(.true.,.true.,pvx_dt,ptemp) 
			!vz time derivative
			ptemp=pcur_xx*scale_lame
			call Dpz_op(.true.,.false.,pvz_dt,ptemp) 
			ptemp=pcur_zz*scale_lame_mu
			call Dpz_op(.true.,.true.,pvz_dt,ptemp) 
			ptemp=pcur_xz*scale_mu_xz
			call Dmx_op(.true.,.true.,pvz_dt,ptemp)
			!sigmaxx derivative
			ptemp=pcur_vx*scalevx
			call Dmx_op(.true.,.false.,pxx_dt,ptemp) 
			!sigmazz derivative
			ptemp=pcur_vz*scalevz
			call Dmz_op(.true.,.false.,pzz_dt,ptemp) 
			!sigmaxz time derivative
			ptemp=pcur_vx*scalevx
			call Dpz_op(.true.,.false.,pxz_dt,ptemp) 
			ptemp=pcur_vz*scalevz
			call Dpx_op(.true.,.true.,pxz_dt,ptemp) 
			
			
			
		else	     !Forward
			!vx time derivative
			call Dmx_op(.false.,.false.,pcur_xx,pvx_dt) 
			call Dpz_op(.false.,.true.,pcur_xz,pvx_dt) 
			pvx_dt = pvx_dt*scalevx
			!vz time derivative
			call Dmz_op(.false.,.false.,pcur_zz,pvz_dt) 
			call Dpx_op(.false.,.true.,pcur_xz,pvz_dt) 
			pvz_dt = pvz_dt*scalevz
			!sigmaxx and sigmazz time derivative
			call Dpx_op(.false.,.false.,pcur_vx,ptemp) 
			pxx_dt = ptemp*scale_lame_mu
			pzz_dt = ptemp*scale_lame
			call Dpz_op(.false.,.false.,pcur_vz,ptemp) 
			pxx_dt = pxx_dt + ptemp*scale_lame
			pzz_dt = pzz_dt + ptemp*scale_lame_mu
			!sigmaxz time derivative
			call Dmx_op(.false.,.false.,pcur_vz,pxz_dt) 
			call Dmz_op(.false.,.true.,pcur_vx,pxz_dt) 
			pxz_dt = pxz_dt*scale_mu_xz
			
		endif
		
		
		
		
		
		
		
		
  
  
  	end subroutine Prop_elastic_step_op
  
  
  	subroutine Prop_elasti_step_close()
  		call D10pt_close()
  		if(allocated(ptemp))deallocate(ptemp)
  		if(allocated(scale_mu_xz))deallocate(scale_mu_xz)
  		if(allocated(scale_lame))deallocate(scale_lame)
  		if(allocated(scale_lame_mu))deallocate(scale_lame_mu)
  		if(allocated(scalevx))deallocate(scalevx)
  		if(allocated(scalevz))deallocate(scalevz)
  		
  	end subroutine  Prop_elasti_step_close



end module Prop_elastic_step